@@ -30,7 +30,7 @@ from swift.common.http import HTTP_OK, HTTP_CREATED, HTTP_ACCEPTED, \
     HTTP_PARTIAL_CONTENT, HTTP_NOT_MODIFIED, HTTP_PRECONDITION_FAILED, \
     HTTP_REQUESTED_RANGE_NOT_SATISFIABLE, HTTP_LENGTH_REQUIRED, \
     HTTP_BAD_REQUEST, HTTP_REQUEST_TIMEOUT, HTTP_SERVICE_UNAVAILABLE, \
-    is_success
+    HTTP_CLIENT_CLOSED_REQUEST, is_success
 
 from swift.common.constraints import check_utf8
 from swift.proxy.controllers.base import get_container_info, \
@@ -87,6 +87,8 @@ def _header_strip(value):
         # behave as though it wasn't provided
         return None
     return stripped
+
+
 _header_strip.re = re.compile('^[\x00-\x20]*|[\x00-\x20]*$')
 
 
@@ -538,7 +540,10 @@ class Request(swob.Request):
             obj = self.environ['PATH_INFO'][1:] or None
             return self.bucket_in_host, obj
 
-        bucket, obj = self.split_path(0, 2, True)
+        try:
+            bucket, obj = self.split_path(0, 2, True)
+        except ValueError:
+            raise InvalidURI(self.path)
 
         if bucket and not validate_bucket_name(bucket):
             # Ignore GET service case
@@ -789,16 +794,18 @@ class Request(swob.Request):
             split_path(self.headers['X-Amz-Copy-Source'], 1, 2, True)
 
         if (self.container_name == source_container and
-                self.object_name == source_obj and
-                self.headers.get('x-amz-metadata-directive',
-                                 'COPY') == 'COPY'):
-            raise InvalidRequest("This copy request is illegal "
-                                 "because it is trying to copy an "
-                                 "object to itself without "
-                                 "changing the object's metadata, "
-                                 "storage class, website redirect "
-                                 "location or encryption "
-                                 "attributes.")
+                self.object_name == source_obj):
+            if self.headers.get('x-amz-metadata-directive', 'COPY') == 'COPY':
+                raise InvalidRequest("This copy request is illegal "
+                                     "because it is trying to copy an "
+                                     "object to itself without "
+                                     "changing the object's metadata, "
+                                     "storage class, website redirect "
+                                     "location or encryption "
+                                     "attributes.")
+            else:
+                self.environ['swift3.copy_to_itself'] = True
+
         return src_resp
 
     def _canonical_uri(self):
@@ -951,6 +958,11 @@ class Request(swob.Request):
             env['HTTP_X_COPY_FROM'] = env['HTTP_X_AMZ_COPY_SOURCE']
             del env['HTTP_X_AMZ_COPY_SOURCE']
             env['CONTENT_LENGTH'] = '0'
+            # Mitigate a bug where overwriting a SLO did not remove parts.
+            if env.get('swift3.copy_to_itself', False):
+                if query is None:
+                    query = dict()
+                query['multipart-manifest'] = 'get'
 
         if CONF.force_swift_request_proxy_log:
             env['swift.proxy_access_log_made'] = False
@@ -1013,6 +1025,12 @@ class Request(swob.Request):
                     HTTP_NO_CONTENT,
                 ],
             }
+            # If bucket creation succeeds after a timeout,
+            # we have to accept that the container already exists.
+            # We rely on the bucket_db to know if the bucket already
+            # exists or not.
+            if self.bucket_db:
+                code_map['PUT'].append(HTTP_NO_CONTENT)
         else:
             # Swift object access.
             code_map = {
@@ -1177,7 +1195,7 @@ class Request(swob.Request):
                     self.bucket_db.release(container)
             return resp
 
-        if self.bucket_db and method == 'PUT':
+        if self.bucket_db and method == 'PUT' and not obj:
             # Container creation failed, remove reservation
             self.bucket_db.release(container)
 
@@ -1199,6 +1217,8 @@ class Request(swob.Request):
             raise AccessDenied()
         if status == HTTP_SERVICE_UNAVAILABLE:
             raise ServiceUnavailable()
+        if status == HTTP_CLIENT_CLOSED_REQUEST:
+            raise RequestTimeout()
 
         raise InternalError('unexpected status code %d' % status)
 
